import requests
from furthrmind.collection.project import Project
import os
from furthrmind import collection
from typing_extensions import List, Dict


class Furthrmind:
    Experiment = collection.Experiment
    File = collection.File
    FieldData = collection.FieldData
    ResearchItem = collection.ResearchItem
    Sample = collection.Sample
    Unit = collection.Unit
    ComboBoxEntry = collection.ComboBoxEntry
    Field = collection.Field
    DataTable = collection.DataTable
    Group = collection.Group
    Column = collection.Column
    Project = collection.Project
    Category = collection.Category

    def __init__(
        self, host, api_key=None, api_key_file=None, project_id=None, project_name=None
    ):

        if not host.startswith("http"):
            host = f"https://{host}"
        self.host = host
        self.base_url = f"{host}/api2"
        self.session = requests.session()
        self.project_url = None

        assert (
            api_key is not None or api_key_file is not None
        ), "Either api_key or api_key_file must be specified"

        if api_key_file:
            assert os.path.isfile(api_key_file), "Api key file is not a valid file"
            with open(api_key_file, "r") as f:
                api_key = f.read()

        self.session.headers.update({"X-API-KEY": api_key})

        self.api_key = api_key
        self.project_id = project_id
        self._write_fm_to_base_class()

        self.set_project(project_id, project_name)

    def set_project(self, id=None, name=None):
        if not id and not name:
            return
        if id:
            self.project_id = id
            self.project_url = f"{self.base_url}/projects/{self.project_id}"
            return
        if name:
            projects = Project._get_all()
            for project in projects:
                if project.name.lower() == name.lower():
                    self.project_id = project.id
                    self.project_url = f"{self.base_url}/projects/{self.project_id}"
                    return
            raise ValueError("Project not found")

    def get_project_url(self, project_id=None):
        if not project_id:
            if self.project_url is None:
                raise ValueError("Project URL not set")
            return self.project_url
        else:
            project_url = f"{self.base_url}/projects/{project_id}"
            return project_url

    def _write_fm_to_base_class(self):
        from furthrmind.collection.baseclass import BaseClass

        BaseClass.fm = self

    def send_email(self, mail_to: str, mail_subject: str, mail_body: str):
        url = f"{self.base_url}/send-email"
        data = {
            "mail_to": mail_to,
            "mail_subject": mail_subject,
            "mail_body": mail_body,
        }
        self.session.post(url, json=data)

    def run_custom_script(
        self, script: str, files: List[Dict] = None, config: dict = None
    ):
        """
        Run a custom script on the server webdatacalc server.
        The script must be a valid python script.
        The files is a list of dictionaries with the following keys:
        - name: The name of the file, including the extension
        - content: The content of the file
        - format: 'text' or 'bytes'
        - fileId: The id of the file, if the file is already uploaded to the server
        The config is a dictionary with additional configuration for the script.:
        """
        url = f"{self.base_url}/run-custom-script"

        if not files:
            files = []
        for f in files:
            if f.get("fileId"):
                continue
            assert "name" in f, "Name not specified"
            assert "content" in f, "Content not specified"
            assert f["format"] in ["text", "bytes"], "Invalid format"

        if not config:
            config = {}
        if not config.get("X-API-KEY"):
            config.update({"X-API-KEY": self.api_key})
        data = {"script": script, "files": files, "config": config}
        response = self.session.post(url, json=data)
        if response.status_code != 200:
            message = response.text
            raise ValueError(f"Server returned an error: {message}")
        return response.json()
        



from inspect import isclass

from typing_extensions import List, Self, Dict, TYPE_CHECKING

from furthrmind.collection.baseclass import BaseClassWithFieldData, BaseClass

if TYPE_CHECKING:
    from furthrmind.collection import File, FieldData


class ComboBoxEntry(BaseClassWithFieldData):
    """
    ComboBoxEntries represent the available options in a list field

    Attributes
    ----------
    id : str
        id of the column
    name : str
        name of the column
    files : List[File]
        List of files belonging to this comboboxentry. See [File](file.md) for more information.
    fielddata : List[FieldData]
        List of field data belonging to this comboboxentry. See [FieldData](fielddata.md) for more information.
    _fetched : bool
        This is a Boolean attribute indicating whether all attributes have been retrieved from the server or only
        the name and ID are present.
    """

    id = ""
    name = ""
    files: List["File"] = []
    fielddata: List["FieldData"] = []

    _attr_definition = {
        "files": {"class": "File"},
        "fielddata": {"class": "FieldData"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self, project_id=None):
        project_url = ComboBoxEntry.fm.get_project_url(project_id)
        url = f"{project_url}/comboboxentry/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/comboboxentry/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/comboboxentry"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/comboboxentry"
        return url

    @classmethod
    def get(cls, id: str = "", project_id: str = "") -> Self:
        """
        Method to get one combobox entry
        If called on an instance of the class, the id of the instance is used

        Parameters
        ----------
        id : str
            id of the requested comboboxentry
        project_id : str
            Optionally, to get a comboboxentry from another project as the furthrmind sdk was initiated with

        Returns
        -------
        Self
            Instance of the ComboBoxEntry class

        Raises
        ------
        AssertionError
            If used as class method and id not specified

        """

        if isclass(cls):
            assert id, "id must be specified"

        return cls._get(id, project_id=project_id)

    # noinspection PyMethodOverriding
    @classmethod
    def get_many(cls, ids: List[str] = (), project_id: str = "") -> List[Self]:
        """
        This method is a class method that retrieves multiple comboboxentries belonging to one project.

        Parameters
        ----------
        ids : List[str]
            List with ids.

        project_id : str, optional
            Optionally to get comboboxentries from another project as the furthrmind sdk was initiated with,
            defaults to an empty string.

        Returns
        -------
        List[Self]
            List with instances of experiment class.

        Raises
        ------
        AssertError
            If `ids` list is empty.

        """

        assert ids, "ids must be specified"
        return cls._get_many(ids, project_id=project_id)

    @classmethod
    def get_all(cls, project_id: str = "") -> List[Self]:
        """
        Method to get all comboboxentries belonging to one project.

        Parameters
        ----------
        project_id : str, optional
            Optionally to get comboboxentries from another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            A list containing instances of the comboboxentry class.
        """

        return cls._get_all(project_id=project_id)

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create(cls, name: str, field_name: str = "", field_id: str = "", project_id: str = "") -> Self:
        """
        Method to create a new combobox entry

        Parameters
        ----------
        name: str
            Name of the combobox entry.
        field_name: str, optional
            Name of the field where the combobox entry should belong to. Either the field name or id must be provided.
        field_id: str, optional
            Id of the field where the combobox entry should belong to. Either the field name or id must be provided.
        project_id: str, optional
            Id of the project where the combobox entry should be created. This is only required if the combobox entry needs to
            be created in a different project than the one the SDK was initiated with.

        Returns
        -------
        Self
            Instance of the column comboboxentry class.

        Raises
        ------
        ValueError
            If the name is not specified.
        ValueError
            If neither field_name nor field_id is provided.
        ValueError
            If the field with the given name is not found.
        """

        from furthrmind.collection.field import Field

        if not name:
            raise ValueError("Name must be specified")
        if not field_name and not field_id:
            raise ValueError("Either field_name or field_id must be provided")

        if field_name:
            fields = Field._get_all(project_id)
            for field in fields:
                if field.name.lower() == field_name.lower():
                    field_id = field.id
                    break

            if not field_id:
                raise ValueError("Field with given name not found")

        data = {"name": name, "field": {"id": field_id}}
        id = cls._post(data, project_id)
        data["id"] = id
        return data

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create_many(cls, data_list: List[Dict], project_id: str = "") -> List[Self]:
        """
        Method to create many new combobox entries

        Parameters
        ----------
        data_list : List[Dict]
            List of dictionaries containing the data for creating instances of comboboxentry class. Each dictionary
            should have the following keys:

            - "name" (str): Name of the combobox entry.
            - "field_name" (str, optional): Name of the field where the combobox entry should belong to. Either
                "field_name" or "field_id" must be provided.
            - "field_id" (str, optional): ID of the field where the combobox entry should belong to. Either "field_name"
                or "field_id" must be provided.

        project_id : str, optional
            ID of the project where the combobox entries should be created. If not provided, the items will be created
            in the current project.

        Returns
        -------
        List[Self]
            List of instances of comboboxentry class that were created.

        Raises
        ------
        ValueError
            If any of the following conditions are met:

            - "name" is missing in any of the dictionaries in "data_list".
            - Both "field_name" and "field_id" are missing in any of the dictionaries in "data_list".
            - The provided "field_name" does not match any existing field in the project.

        """

        from furthrmind.collection.field import Field

        look_for_field_ids = False
        for data in data_list:
            if not data.get("name"):
                raise ValueError("Name must be specified")

            if data.get("field_name"):
                look_for_field_ids = True
                break

        if look_for_field_ids:
            fields = Field._get_all(project_id)
            for data in data_list:
                field_name = data.get("field_name")
                field_id = data.get("field_id")
                if not field_name and not field_id:
                    raise ValueError("Either field_name or field_id must be provided")
                if field_name:
                    for field in fields:
                        if field.name.lower() == field_name.lower():
                            field_id = field.id
                            data["field_id"] = field_id
                            break
                        if not data.get("field_id"):
                            raise ValueError(f"Field with given name '{field_name}' not found")

        new_data_list = []
        for data in data_list:
            new_data_list.append({
                "name": data.get("name"),
                "field": {"id": data.get("field_id")}
            })

        id_list = cls._post(new_data_list, project_id, force_list=True)
        for data, id in zip(new_data_list, id_list):
            data["id"] = id

        return new_data_list



from datetime import datetime, date

from bson import ObjectId
from typing_extensions import List, TYPE_CHECKING, Dict, Tuple, Union, Self

from furthrmind.collection.baseclass import BaseClass
from furthrmind.utils import instance_overload

if TYPE_CHECKING:
    from furthrmind.collection.unit import Unit


class FieldData(BaseClass):
    """
    Attributes
    ----------
    id : str
        id of the fielddata
    field_name : str
        field name of the corresponding field
    field_id : List[File]
        field id of the corresponding field
    field_type : str
        field type of the corresponding field
    value : Any
        value of the fielddata. Type depends on the field type:

            - Numeric fields: float or None
            - Numeric range fields: list of two floats or None
            - Date fields: python datetime object or None
            - Text field: string or None
            - List field: The value will be a dictionary with the name and the id to the selected option (comboboxentry).
                If no option is selected, the value will be None
            - Notebook field: The value will be a dictionary with the id and content of the notebook.
            - Checkbox field: boolean
    si_value : Union[float, None]
        In case of numeric fields, the attribute represents the corresponding si-value considering the selected unit
    unit : [Unit](unit.md)
        The selected unit for numeric fields. Otherwise None
    author: Dict[str, str]
        The author of the fielddata with id and email address

    """
    id = ""
    field_name = ""
    field_id = ""
    field_type = ""
    si_value = None
    unit: "Unit" = None
    author = None
    value = None

    _attr_definition = {
        "unit": {"class": "Unit"},
        "field_name": {"data_key": "fieldname"},
        "field_type": {"data_key": "fieldtype"},
        "field_id": {"data_key": "fieldid"},
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

        # create instance methods for certain class_methods
        instance_methods = ["_check_value_type"]
        instance_overload(self, instance_methods)

    def _update_attributes(self, data):
        super()._update_attributes(data)
        if self.field_type == "Date":
            if self.value:
                self.value = datetime.fromtimestamp(self.value)

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/fielddata"
        return url

    @classmethod
    def get(cls, id=None):
        raise TypeError("Not implemented")

    @classmethod
    def _get_all(cls, project_id=None):
        raise TypeError("Not implemented")

    def update_value(self, value) -> str:
        """
        Parameters
        ----------
        value :
            The value to update the fielddata. The valid types for each field type are as follows:

            - numeric, numeric-field, numeric_field: float or int, or a string convertable to a float
            - numeric range, numeric-range-field, numeric_range_field: List with two floats, ints, or strings convertable to a float
            - date, date_field, date-field, datefield: datetime, or date object, or unix timestamp or string with
                iso format
            - singleline, singlelinefield, text, text-field, text_field, textfield: string
            - combobox, comboboxfield, list, list-field, list_field, listfield: dict with id or name as key, or string
                with name, or string with id
            - multiline, notebook, notebookfield, notebook-field, notebook_field: dict with content as key, or string
            - checkbox, checkbox-field, checkbox_field, checkboxfield: boolean

        Returns
        -------
        id : str
            The id of the updated fielddata.
        """

        value, field_type = self.__class__._check_value_type(value, self.field_type)
        data = {"id": self.id,
                "value": value}
        id = self._post(data)
        self.value = value
        return id

    def set_calculation_result(self, value: Dict) -> str:
        """
        Parameters
        ----------
        value: dict
            A dictionary representing the calculation result.

        Raises
        ------
        TypeError
            If the field type is not a calculation

        Returns
        -------
        id: str
            The ID of the calculation field.
        """

        if not self.field_type.lower() in ["calculation", "rawdatacalc"]:
            raise TypeError("Only applicable for calculation field")

        url = f"{self.fm.base_url}/set-result/{self.id}"
        response = self.fm.session.post(url, json=value)
        if response.status_code != 200:
            raise ValueError("Setting calculation result failed")
        return self.id

    @classmethod
    def _check_value_type(cls, value, field_type=None) -> Tuple:
        from furthrmind.collection import Field

        if issubclass(cls, BaseClass):
            # classmethod
            if field_type is None:
                raise ValueError("fieldtype must not be None")
        else:
            # instance method
            self = cls
            field_type = self.field_type

        # raises an error or returns: "Numeric", "Date", "SingleLine",
        # "ComboBox", "MultiLine", "CheckBox", "Calculation"
        field_type = Field._check_field_type(field_type)

        if value is None:
            return value, field_type

        if field_type == "Numeric":
            try:
                value = float(value)
            except:
                raise TypeError("Not numeric")
            return value, field_type
        elif field_type == "NumericRange":
            if not isinstance(value, list):
                raise TypeError("Not a list")
            if not len(value) == 2:
                raise TypeError("Not a list with two elements")
            for pos, element in enumerate(value):
                if element is None:
                    continue
                try:
                    element = float(element)
                    value[pos] = element
                except:
                    raise TypeError("Element is not convertible to float")
            return value, field_type

        elif field_type == "Date":
            if isinstance(value, datetime):
                if value.tzinfo is None:
                    timezone = value.astimezone().tzinfo
                    value = value.replace(tzinfo=timezone)
                return int(value.timestamp()), field_type
            if isinstance(value, date):
                value = datetime.combine(value, datetime.min.time())
                return int(value.timestamp()), field_type
            if isinstance(value, str):
                try:
                    value = datetime.fromisoformat(value)
                    return int(value.timestamp()), field_type
                except ValueError:
                    raise TypeError("No iso time format")
            if isinstance(value, (int, float)):
                return value, field_type
        elif field_type == "SingleLine":
            if isinstance(value, str):
                return value, field_type
            if isinstance(value, (float, int)):
                return str(value), field_type
            raise TypeError("Type must be string")

        elif field_type == "ComboBox":
            if isinstance(value, dict):
                if "id" in value:
                    return value, field_type
                if "name" in value:
                    return value, field_type
                raise TypeError("The dict must have either id or name key")
            if isinstance(value, str):
                try:
                    value = ObjectId(value)
                    value = {"id": value}
                except:
                    value = {"name": value}
                return value, field_type
            raise TypeError("Only string and dict supported")

        elif field_type == "MultiLine":
            if isinstance(value, dict):
                if "content" not in value:
                    raise TypeError("Key 'content' is required")
                return value, field_type
            if isinstance(value, str):
                value = {"content": value}
                return value, field_type
            raise TypeError("Only string and dict supported")

        elif field_type == "CheckBox":
            if not isinstance(value, bool):
                raise TypeError("value must be a bool")
            return value, field_type
        elif field_type == "Calculation":
            return None, field_type

    def update_unit(self, unit: Union[Dict, str]) -> str:
        """
        Parameters
        ----------
        unit : Union[Dict, str]
            Dictionary with id or name, or string representing the name, or string representing the id.

        Returns
        -------
        id: str
            The id of the updated unit.

        Raises
        ------
        None

        """

        unit = self._check_unit(unit)
        data = {"id": self.id,
                "unit": unit}
        id = self._post(data)
        self.unit = unit
        return id

    @classmethod
    def _check_unit(cls, unit):
        if not unit:
            return unit
        if isinstance(unit, dict):
            if "id" in unit:
                return unit
            if "name" in unit:
                return unit
            raise TypeError("The dict must have either id or name key")

        elif isinstance(unit, str):
            try:
                unit = ObjectId(unit)
                unit = {"id": unit}
            except:
                unit = {"name": unit}
            return unit
        raise TypeError("Only string and dict supported")

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create(cls, field_name: str = "", field_type: str = "", field_id: str = "", value=None, unit=None,
               project_id: str = "") -> Self:
        """
        Parameters
        ----------
        field_name : str, optional
            Name of the field. Either field name and field_type must be specified, or field_id must be specified.
        field_type : str, optional
            Type of the field. Must be one of the following:

                - Numeric fields: numeric, numeric-field, numeric_field
                - Numeric range fields: numericrange, numeric_range, numericrangefield, numeric-range-field, numeric_range_field
                - Date fields: date, date_field, date-field, datefield
                - Text fields: singleline, singlelinefield, text, text-field, text_field, textfield
                - List fields: combobox, comboboxfield, list, list-field, list_field, listfield
                - Notebook fields: multiline, notebook, notebookfield, notebook-field, notebook_field
                - Checkbox fields: checkbox, checkbox-field, checkbox_field, checkboxfield
                - Calculation fields: calculation, calculation-field, calculation_field, calculationfield

        field_id : str, optional
            ID of the field.
        value : None, float, int, str, datetime, date, optional
            Value of the field. The data type depends on the field_type parameter:

                - Numeric fields: float or int, or a string convertible to a float
                - Numeric range fields: List with two floats, ints, or strings convertable to a float
                - Date fields: datetime, date object, unix timestamp, or string with iso format
                - Text fields: string
                - List fields: dictionary with id or name as key, or string with name, or string with id
                - Notebook fields: dictionary with content as key, or string
                - Checkbox fields:  boolean

        unit : dict, str, optional
            Unit of the field. Can be either a dictionary with id or name, or a string with the name.
        project_id : str, optional
            Optionally to create fielddata in another project as the furthrmind sdk was initiated with

        Returns
        -------
        dict
            Instance of the `fielddata` class.

        Raises
        ------
        ValueError
            If field_id not specified, fieldname and fieldtype must be specified.

        """

        from furthrmind.collection import Field
        data = {}
        if field_id:
            data.update({"fieldid": field_id})
            field: Field = Field.get(id=field_id)
            field_type = field.type

        value, field_type = FieldData._check_value_type(value, field_type)
        data["value"] = value

        if not field_id:
            if not field_name or not field_type:
                raise ValueError("fieldname and fieldtype must be specified")
            data.update(
                {"fieldname": field_name, "fieldtype": field_type})

        if unit:
            unit = FieldData._check_unit(unit)
            data["unit"] = unit

        id = FieldData._post(data, project_id)
        data["id"] = id
        return data

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create_many(cls, data_list: List[Dict], project_id: str = "") -> List[Self]:
        """
        Parameters
        ----------
        data_list: List[Dict]
            List with dictionaries containing the following keys:
            - field_name: name of the field. Either field name and field_type must be specified, or field_id must be specified
            - field_type: type of the field. Must be one of the following:

                - Numeric fields: numeric, numeric-field, numeric_field
                - Numeric range fields: numericrange, numeric_range, numericrangefield, numeric-range-field, numeric_range_field
                - Date fields: date, date_field, date-field, datefield
                - Text fields: singleline, singlelinefield, text, text-field, text_field, textfield
                - List fields: combobox, comboboxfield, list, list-field, list_field, listfield
                - Notebook fields: multiline, notebook, notebookfield, notebook-field, notebook_field
                - Checkbox fields: checkbox, checkbox-field, checkbox_field, checkboxfield
                - Calculation fields: calculation, calculation-field, calculation_field, calculationfield

            - field_id: id of the field
            - value:

                - Numeric fields: float or int, or a string convertible to a float
                - Numeric range fields: List with two floats, ints, or strings convertable to a float
                - Date fields: datetime, date object, unix timestamp, or string with iso format
                - Text fields: string
                - List fields: dictionary with id or name as key, or string with name, or string with id
                - Notebook fields: dictionary with content as key, or string
                - Checkbox fields:  boolean

            - unit: dictionary with id or name, or name as string, or id as string

        project_id: str, optional
            Optionally to create an item in another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Dict]
            List with dictionaries of instances of the fielddata class

        """

        from furthrmind.collection import Field

        post_data_list = []
        for data in data_list:
            field_id = data.get("field_id")
            field_name = data.get("field_name")
            field_type = data.get("field_type")
            value = data.get("value")
            unit = data.get("unit")

            _data = {}
            if field_id:
                _data.update({"fieldid": field_id})
                field: Field = Field.get(id=field_id)
                field_type = field.type

            value, field_type = FieldData._check_value_type(value, field_type)
            _data["value"] = value

            if not field_id:
                if not field_name or not field_type:
                    raise ValueError("field_name and field_type must be specified")
                _data.update(
                    {"fieldname": field_name,
                     "fieldtype": field_type})

            if unit:
                unit = FieldData._check_unit(unit)
                _data["unit"] = unit
            post_data_list.append(_data)

        id_list = FieldData._post(post_data_list, project_id, force_list=True)
        for data, id in zip(post_data_list, id_list):
            data["id"] = id
        return post_data_list



from furthrmind.collection.baseclass import BaseClassWithFieldData, BaseClass
from typing_extensions import List, Dict, Self, TYPE_CHECKING
from inspect import isclass
if TYPE_CHECKING:
    from furthrmind.collection import File, FieldData, Experiment, Sample, ResearchItem


class Group(BaseClassWithFieldData):
    """
    Attributes
    ----------
    id : str
        id of the group
    name : str
        name of the group
    shortid : str
        shortid of the group
    files : List[File]
        List of files belonging to this group. See [File](file.md) for more information.
    fielddata : List[FieldData]
        List of field data belonging to this group. See [FieldData](fielddata.md) for more information.
    samples : List[Sample]
        This list contains 'sample' objects belonging to this group. These objects are partially fetched,
        providing only the name and ID. To retrieve the entire object, invoke the 'get()' method on the 'sample'.
        Refer to nested objects in [Getting Started](index.md) for further details. For a comprehensive understanding of
        'sample' objects, see the provided [Sample](sample.md).
    experiments : List[Experiment]
        This list contains 'experiment' objects belonging to this group. These objects are partially fetched,
        providing only the name and ID. To retrieve the entire object, invoke the 'get()' method on the 'experiment'.
        Refer to nested objects in [Getting Started](index.md) for further details. For a comprehensive understanding of
        'experiment' objects, see the provided [Experiment](experiment.md).
    researchitems : Dict[str, List[ResearchItem]]
        This is a dictionary with category name as keys and lists with the corresponding `researchitem` objects belonging
         to this group as values. The `researchitem` objects are partially fetched,
        providing only the name and ID. To retrieve the entire object, invoke the 'get()' method on the 'researchitem'.
        Refer to nested objects in [Getting Started](index.md) for further details. For a comprehensive understanding of
        'researchitem' objects, see the provided [ResearchItem](researchitem.md).
    sub_groups : List[Group]
        This list contains 'group' objects that are subgroups of this group. These objects are partially fetched,
        providing only the name and ID. To retrieve the entire object, invoke the 'get()' method on the 'group'.
        Refer to nested objects in [Getting Started](index.md) for further details. For a comprehensive understanding of
        'group' objects, see the provided [Group](group.md).
    parent_group : Group
        If the group is a subgroup, the attribute holds its parent group. This object is partially fetched,
        providing only the name and ID. To retrieve the entire object, invoke the 'get()' method on the 'group'.
        Refer to nested objects in [Getting Started](index.md) for further details. For a comprehensive understanding of
        'group' objects, see the provided [Group](group.md).
    _fetched : bool
        This is a Boolean attribute indicating whether all attributes have been retrieved from the server or only
        the name and ID are present.
    """

    id = ""
    name = ""
    neglect = False
    shortid = ""
    files: List["File"] = []
    fielddata: List["FieldData"] = []
    experiments: List["Experiment"] = []
    samples: List["Sample"] = []
    researchitems: Dict[str, List["ResearchItem"]] = {}
    sub_groups: List[Self] = []
    parent_group: Self = None

    _attr_definition = {
        "files": {"class": "File"},
        "fielddata": {"class": "FieldData"},
        "samples": {"class": "Sample"},
        "experiments": {"class": "Experiment"},
        "researchitems": {"class": "ResearchItem", "nested_dict": True},
        "sub_groups": {"class": "Group"},
        "parent_group": {"class": "Group"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self, project_id=None):
        project_url = Group.fm.get_project_url(project_id)
        url = f"{project_url}/groups/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/groups/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/groups"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/groups"
        return url

    @classmethod
    def get(cls, id: str = "", name: str = "", shortid: str = "", parent_group_id: str = "", project_id: str = "") -> Self:
        """
        Method to get one group by its id, name, or shortid.
        If called on an instance of the class, the id of the class is used.

        Parameters
        ----------
        id : str
            id or short_id of requested group
        name : str
            name of requested group. For subgroups, see parent_group_id parameter
        shortid : str
            shortid of requested group
        parent_group_id : str
            id of parent group.
            If a subgroup is requested, the name and the parent_group_id is required
        project_id : str, optional
            Optionally to get a group from another project as the furthrmind sdk was initiated with.
            Defaults to None

        Returns
        -------
        Self
            Instance of group class

        Raises
        ------
        AssertionError
            When used as a class method, id, name, or shortid must be specified.

        """

        if isclass(cls):
            assert id or name or shortid, "Either id or name must be specified"

        return cls._get(id, shortid, name, parent_group_id=parent_group_id, project_id=project_id)

    @classmethod
    def get_many(cls, ids: List[str] = (), shortids: List[str] = (), names: List[str] = (),
                 project_id: str = "") -> List[Self]:
        """
        Method to get many groups

        Parameters
        ----------
        ids : List[str]
            List of ids to filter the groups.
        shortids : List[str]
            List of short_ids to filter the groups.
        names : List[str]
            List of names to filter the groups.
        project_id : str, optional
            Optionally to get groups from another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            List of instances of the Group class.

        Raises
        ------
        AssertionError
            If none of the parameters (ids, shortids, or names) are specified.

        """

        assert ids or names or shortids, "Either ids, shortids, or names must be specified"
        return cls._get_many(ids, shortids, names, project_id=project_id)

    @classmethod
    def get_all(cls, project_id: str = "") -> List[Self]:
        """
        Method to get all groups belonging to one project

        Parameters
        ----------
        project_id : str, optional
            Optionally to get groups from another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            A list of instances of the group class.

        """

        return super()._get_all(project_id)
    
    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create(cls, name: str, parent_group_id: str = "", project_id: str = "") -> Self:
        """
        Method to create a new group.

        Parameters
        ----------
        name: str
            The name of the item to be created.
        parent_group_id: str, optional
            The ID of the parent group where the new group should belong to. Defaults to an empty string.
        project_id: str, optional
            Optionally, create a group in another project as the furthrmind SDK was initiated with

        Returns
        -------
        instance of the group class
        """

        data = {"name": name}
        if parent_group_id:
            data["parent_group"] = {"id": parent_group_id}
        id = cls._post(data, project_id)
        data["id"] = id
        return data

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create_many(cls, name_list: List[str], project_id: str = "") -> List[Self]:
        """
        Method to create multiple groups

        Parameters
        ----------
        name_list: List[str]
            A list containing names of the groups to be created.
        project_id: str, optional
            Optionally to create groups in another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            A list containing instances of the group class corresponding to the groups created.

        """

        data_list = [{"name": name} for name in name_list]

        id_list = cls._post(data_list, project_id, force_list=True)

        for data, id in zip(data_list, id_list):
            data["id"] = id

        return data_list



from furthrmind.collection.baseclass import BaseClass
import os
from io import BytesIO

class File(BaseClass):
    """
    Attributes
    ----------
    id : str
        id of the file
    name : str
        name of the file including the extension
    """

    id = ""
    name = ""

    _attr_definition = {
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    @classmethod
    def get(cls, id=None):
        raise TypeError("Not implemented")

    @classmethod
    def _get_all(cls):
        raise TypeError("Not implemented")

    def download(self, folder: str, overwrite: bool = False):
        """
        Method to download a file

        Parameters
        ----------
        folder : str
            The folder where the file should be saved
        overwrite : bool, optional
            Whether to overwrite the existing file in the folder if it already exists (default is False)
        """

        from furthrmind.file_loader import FileLoader
        fl = FileLoader(self.fm.host, self.fm.api_key)

        if not os.path.isdir(folder):
            raise ValueError("Folder does not exist")
        fl.downloadFile(self.id, folder, overwrite)

    def download_bytes(self) -> BytesIO:
        """
        Method to download a file and save to BytesIO object

        Returns:
            BytesIO: The downloaded file stored as BytesIO object
        """

        from furthrmind.file_loader import FileLoader
        fl = FileLoader(self.fm.host, self.fm.api_key)

        flag, bytes_io = fl.downloadFile(self.id, bytesIO=True)
        return bytes_io

    def update_file(self, file_path: str, file_name: str = ""):
        """
        Update a file.

        Parameters
        ----------
        file_path : str
            The path to the file.
        file_name : str, optional
            The new file name. Defaults to "". If not set, the file_name is taken from the file_path.

        Raises
        ------
        ValueError
            If the file does not exist.

        """
        from furthrmind.file_loader import FileLoader
        fl = FileLoader(self.fm.host, self.fm.api_key)

        if not os.path.isfile(file_path):
            raise ValueError("File does not exist")

        fl.updateFile(self.id, file_path, file_name)




